#include "uart_interpreter.h"
#include "uart.h"
#include "thread_coap_utils.h"
#include "app_scheduler.h"

typedef enum
{
    FRAME_SYNC_FLAG,
    FRAME_PAYLOAD_LENGTH,
    FRAME_CMD_PART_HIGH,
    FRAME_CMD_PART_LOW,
    FRAME_DATA,
} frame_part;

/**@brief Structure storing information about current test frame payload. */
typedef union
{
    uint8_t  frame[UART_RX_BUF_SIZE];
    uint32_t current_frame_number;
} frame_payload_t;

/**@brief Structure storing information about current test frame used in sending acknowledgments. */
typedef struct
{
    uint16_t          length;
    uint16_t          payload_length;
    uint8_t  * p_frame;
} frame_information_t;

static frame_payload_t     m_frame_payload;
static frame_information_t m_frame =
{
    .length = 0,
    .payload_length = 0,
    .p_frame      = &m_frame_payload.frame,
};

static frame_part sm_frame_part;

void uart_interpreter_process_data_received(uint8_t data)
{
    static uint8_t payload_length = 0;
    static uint16_t cmd = 0;
    static uint8_t *parameters;
    static uint8_t index_parameters = 0;
    static uint8_t nbr_parameters = 0;
    static uint8_t frameIndex = 0;
    uint32_t err_code = NRF_SUCCESS;

    

    switch(sm_frame_part)
    {
        case FRAME_SYNC_FLAG:
            if (data == 0xAA)
            {
                sm_frame_part = FRAME_PAYLOAD_LENGTH;
                free(parameters);
                frameIndex=0;
            }
            //To to: start a timer to detect timeout
            break;

        case FRAME_PAYLOAD_LENGTH:
            payload_length = data;
            m_frame.length = data+2;
            parameters = (uint8_t *) malloc(payload_length);
            sm_frame_part = FRAME_CMD_PART_HIGH;
            break;

        case FRAME_CMD_PART_HIGH:
            cmd = data;
            //m_frame.p_frame[payloadIndex] = data;
            cmd = cmd<<8;
            sm_frame_part = FRAME_CMD_PART_LOW;
            break;

        case FRAME_CMD_PART_LOW:
            cmd += data;
            //m_frame.p_frame[payloadIndex] = data;
            switch(cmd)  //To do : replace the numbers by defines
            {
                case 1:  //CMD_TEST_MAIN_UART
                    nbr_parameters = 0;
                    break;

                case 2:  //CMD_TEST_LOG
                    nbr_parameters = 0;
                    break;

                case 3:  //CMD_LIST_IPV6_ADDRESSES_OF_THREAD_INTERFACE
                    nbr_parameters = 0;
                    break;

                case 4:  //CMD_GET_RLOC_ADDRESS
                    nbr_parameters = 0;
                    break;

                case 5:  //CMD_GET_NEIGBORS
                    nbr_parameters = 0;
                    break;

                case 6:  //CMD_GET_EID_CACHE:
                nbr_parameters = 0;
                    break;

                case 7:  //CMD_SEND_FRAME_TO_SPECIFIC_DEVICE
                    nbr_parameters = 0;
                    break;

               case 8:  //CMD_SEND_FRAME_IN_MULTICAST
                    nbr_parameters = 0;  //To do : set the correct value -------------------------------
                    break;

                default:
                    nbr_parameters = 0;
                    break;
            }
            sm_frame_part = FRAME_DATA;


        case FRAME_DATA:
            if(index_parameters<nbr_parameters)
            {
                parameters[index_parameters] = data;
                //m_frame.p_frame[payloadIndex] = data;
                index_parameters++;
            }
            
            if((index_parameters+2)>=payload_length)
            {
                send_ack(0xAA, cmd, CMD_RESULT_OK, 0, NULL);

                //process_uart_cmd(cmd, parameters); // To do: use the scheduler
                err_code = app_sched_event_put(&m_frame, sizeof(m_frame), process_received_frame);
                ASSERT(err_code == NRF_SUCCESS);

                sm_frame_part = FRAME_SYNC_FLAG;
            }
            break;

        //To do : add checksum and check it to process the frame

        default:
            sm_frame_part = FRAME_SYNC_FLAG;
            break;
    }

    m_frame.p_frame[frameIndex] = data;

    frameIndex++;
}

/**@brief This function is used to process the reveived frame.
 *
 * @param[in]    p_event_data Pointer to the event data.
 * @param[in]    event_size   Size of the event.
 */
static void process_received_frame(void * p_event_data, uint16_t event_size)
{
    uint32_t err_code = NRF_SUCCESS;

    uart_log_send_stringWithFrame("Frame received: ", m_frame.p_frame, m_frame.length);

    if(m_frame.length>=4)
    {
        uint16_t cmd = m_frame.p_frame[2];
        cmd = cmd<<8;
        cmd = cmd|(m_frame.p_frame[3]&0xFF);
        uint8_t *parameters = malloc(m_frame.length-4);
        for(int i=4; i<event_size; i++)
        {
            parameters[i-4] = m_frame.p_frame[i];
        }
        process_uart_cmd(cmd, parameters);
        free(parameters);
    }

    
    /*NRF_LOG_DEBUG("Ready to send data over thread network");
    do
    {
        uint16_t length = (uint16_t) m_frame.payload_length;
        thread_command_handler(m_frame.p_frame, length);
    } while (err_code == NRF_ERROR_BUSY);      
    nrf_serial_rx_drain(&serial0_uarte);*/
}

static void process_uart_cmd(uint16_t cmd, uint8_t *cmd_parameters)
{
    ret_code_t ret;
    const char  message_unknownCmd[] = "Unknown cmd received\r\n";
    const char  message_test[] = "Test\r\n";
    char  u_message_test[] = "Ceci est un test\r\n";

    switch(cmd)
    {
        case 1:  // Test main uart
            ret = uart_log_send_string(message_test);
            break;

        case 2:  // Test cmd
            ret = uart_log_send_string(message_test);
            break;

        case 3:  // List all IPv6 addresses assigned to the Thread interface
            ret = get_list_ip_thread_interface();
            break;

        case 4:  // Get our RLOC address (Thread Routing Locator)
            ret = get_thread_rloc_address();
            break;

        case 5:  // Get neighbors informations
            ret = get_thread_neighbors_list();
            break;

        case 6:  // Get the routing table
            clear_routingTable();
            ret = get_thread_routing_table();
            break;

        case 7:  //Send test message to specific device
            ret = send_thread_message(0, u_message_test, strlen(u_message_test));
            break;

        case 8:
            thread_coap_utils_multicast_uart_request_send(cmd_parameters,
                                                           sizeof(cmd_parameters),
                                                           THREAD_COAP_UTILS_MULTICAST_REALM_LOCAL);
            break;

        default:
            ret = uart_log_send_string(message_unknownCmd);
            break;
    }

    APP_ERROR_CHECK(ret);
}

ret_code_t get_list_ip_thread_interface()
{
    ret_code_t ret = 0;

    thread_coap_utils_get_list_ip_thread_interface();

    return ret;
}

ret_code_t get_thread_rloc_address()
{
    ret_code_t ret = 0;

    thread_coap_utils_get_rloc_address();
    otIp6Address  *rlocAddress = thread_get_rlocAddress();

    uart_log_send_stringWithFrame("Rloc address: ", rlocAddress, 8);  

    return ret;
}

ret_code_t get_thread_neighbors_list()
{
    ret_code_t ret = 0;

    thread_coap_utils_get_neighbors_list();
    thread_neighbors list_neighbors = thread_list_neighbors_get();

    char string_array[5];

    for(int i=0; i<list_neighbors.nbr_neighbors; i++)
    {
      char role[8] = "Role: ";
      role[6] = list_neighbors.neighbors[i].mIsChild  ? 'C' : 'R';
      uart_log_send_string(&role); 

      itoa(list_neighbors.neighbors[i].mRloc16, string_array, 10);
      uart_log_send_stringWithFrame("mRloc16: ", string_array, 2);  

      uart_log_send_stringWithFrame("macExtendedAddress: ", list_neighbors.neighbors[i].mExtAddress.m8, 8);        
    }

    return ret;
}

ret_code_t get_thread_routing_table()
{
    ret_code_t ret = 0;

    thread_coap_utils_generate_routing_table();

    return ret;
}

ret_code_t send_thread_message(uint8_t indexInRoutingTable, char* message, uint8_t messageLength)
{
    ret_code_t ret = 0;
    routing_table_t *routingTable;
    otIp6Address *destAddress;

    routingTable = get_routing_table();
    destAddress = routingTable->device[0]->ip6Address; //Send a message to the first IP of the table
    #ifdef DEBUG
        uart_log_send_stringWithFrame("Send message to: ",  destAddress->mFields.m8, sizeof(destAddress->mFields.m8));
    #endif

    thread_coap_utils_unicast_uart_request_send(destAddress, messageLength, message); 


    return ret;
}

ret_code_t send_ack(uint8_t syncFlag, uint16_t cmd, bool cmdResult, uint8_t nbrData, uint8_t *data)
{
	ret_code_t  ret;
        uint8_t *data_array = malloc(4+nbrData);

        data_array[0] = syncFlag;
        data_array[1] = sizeof(data_array)-2;

        uint16_t ackCmd = cmd;

        if(cmdResult == CMD_RESULT_OK)
        {
            ackCmd = ackCmd|0x8000;
        }
        else
        {
            ackCmd = ackCmd|0xC000;
        }

        data_array[2] = (ackCmd&0xFF00)>>8;
        data_array[3] = (ackCmd&0xFF);

        for(int i=0; i<nbrData; i++)
        {
            data_array[4+i] = data[i];
        }

        ret = uart_main_send_bytesArray(data_array, sizeof(data_array));

        free(data_array);

	return ret;
}